/*
 * yasapi_player.c
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
// Rendering a Stream
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd316756%28v=vs.85%29.aspx

// Exclusive-Mode Streams
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd370844%28v=vs.85%29.aspx

#if ! defined (AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY) // {
#define AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY 0x08000000
#endif // }

#if ! defined(AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM) // {
#define AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM 0x80000000
#endif // }

///////////////////////////////////////////////////////////////////////////////
int PlayerCreate(Player *pPlayer, HINSTANCE hModule, const wchar_t *path)
{
  char dummy[MAX_PATH];
  int nLen;

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(pPlayer,sizeof *pPlayer);
  pPlayer->hDlgConfig=NULL;

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->options.path=path;
  OptionsCommonLoad(&pPlayer->options.common,L"    ",path);
  DPUTS(0,"  common options loaded\n");
  OptionsDeviceLoad(&pPlayer->options.device,L"    ",
      pPlayer->options.common.szId,path);
  DPUTS(0,"  device options loaded\n");

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->state=PLAYER_STATE_NULL;
  pPlayer->base.qVolume=1.0;

  /////////////////////////////////////////////////////////////////////////////
  nLen=GetModuleFileNameA(
    hModule,    // _In_opt_ HMODULE hModule,
    dummy,      // _Out_    LPTSTR  lpFilename,
    MAX_PATH-1  // _In_     DWORD   nSize
  );

  if (ERROR_INSUFFICIENT_BUFFER==GetLastError()) {
    DMESSAGEV("getting module name (%s)",YASAPI_MODULE);
    goto module;
  }

  nLen=strlen(dummy);

  if (NULL==(pPlayer->base.pszModuleName=malloc(nLen+1))) {
    DMESSAGE("allocating buffer for module name");
    goto malloc;
  }

  strcpy(pPlayer->base.pszModuleName,dummy);
  pPlayer->base.pszFileName=pPlayer->base.pszModuleName;
  pPlayer->base.pszFileName+=strlen(pPlayer->base.pszFileName);

  while (pPlayer->base.pszModuleName<pPlayer->base.pszFileName
      &&'/'!=pPlayer->base.pszFileName[-1]
      &&'\\'!=pPlayer->base.pszFileName[-1]
      &&':'!=pPlayer->base.pszFileName[-1]) {
    --pPlayer->base.pszFileName;
  }

  DPRINTF(0,"  got module name: \"%s\"\n",pPlayer->base.pszFileName);
  pPlayer->base.hModule=hModule;

  /////////////////////////////////////////////////////////////////////////////
  if (QueueCreate(&pPlayer->base.queue)<0) {
    DMESSAGE("creating queue");
    goto queue;
  }

  DPUTS(0,"  queue created\n");

  /////////////////////////////////////////////////////////////////////////////
  if (StoreCreate(&pPlayer->base.store)<0) {
    DMESSAGE("creating store");
    goto store;
  }

  DPUTS(0,"  store created\n");

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->base.hEvent=CreateEvent(
    0,        // _In_opt_ LPSECURITY_ATTRIBUTES lpEventAttributes,
    FALSE,    // _In_     BOOL                  bManualReset,
    FALSE,    // _In_     BOOL                  bInitialState,
    NULL      // _In_opt_ LPCTSTR               lpName
  );

  if (NULL==pPlayer->base.hEvent) {
    DMESSAGE("creating event");
    goto event;
  }
  
  DPUTS(0,"  event created\n");

  // create the timer /////////////////////////////////////////////////////////
  pPlayer->base.timer.hRead=CreateWaitableTimer(
    NULL,       // _In_opt_  LPSECURITY_ATTRIBUTES lpTimerAttributes,
    FALSE,      // _In_      BOOL bManualReset,
    NULL        // _In_opt_  LPCTSTR lpTimerName
  );

  if (NULL==pPlayer->base.timer.hRead) {
    DMESSAGE("creating timer for triggering read");
    goto timer_read;
  }

  DPUTS(0,"  waitable timer for triggering read created\n");

#if defined (YASAPI_CHECK_UNDERFLOW) // {
  // create the timer /////////////////////////////////////////////////////////
  pPlayer->base.timer.hUnderflow=CreateWaitableTimer(
    NULL,       // _In_opt_  LPSECURITY_ATTRIBUTES lpTimerAttributes,
    FALSE,      // _In_      BOOL bManualReset,
    NULL        // _In_opt_  LPCTSTR lpTimerName
  );

  if (NULL==pPlayer->base.timer.hUnderflow) {
    DMESSAGE("creating timer for triggering check for underflow");
    goto timer_underflow;
  }

  DPUTS(0,"  waitable timer for triggering check underflow created\n");
#endif // }

  // create the thread ////////////////////////////////////////////////////////
  pPlayer->base.hThread=CreateThread(
    NULL,           // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
    0,              // _In_       SIZE_T dwStackSize,
    PlayerThread,   // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
    pPlayer,        // _In_opt_   LPVOID lpParameter,
    0,              // _In_       DWORD dwCreationFlags,
    NULL            // _Out_opt_  LPDWORD lpThreadId
  );

  if (NULL==pPlayer->base.hThread) {
    DMESSAGE("creating thread");
    goto thread;
  }

  DPUTS(0,"  thread created\n");

  /////////////////////////////////////////////////////////////////////////////
  ++pPlayer->state;

  /////////////////////////////////////////////////////////////////////////////
  if (PlayerSend(pPlayer,0,PlayerRun)<0) {
    DMESSAGE("running");
    goto run;
  }

  DPUTS(0,"  run request send\n");

  return 0;
// cleanup:
  //PlayerSend(pPlayer,0,PlayerStop);
run:
  CloseHandle(pPlayer->base.hThread);
thread:
#if defined (YASAPI_CHECK_UNDERFLOW) // {
  CloseHandle(pPlayer->base.timer.hUnderflow);
timer_underflow:
#endif // }
  CloseHandle(pPlayer->base.timer.hRead);
timer_read:
  CloseHandle(pPlayer->base.hEvent);
event:
  StoreDestroy(&pPlayer->base.store);
store:
  QueueDestroy(&pPlayer->base.queue);
queue:
  free(pPlayer->base.pszModuleName);
malloc:
module:
  return 0;
}

void PlayerDestroy(Player *pPlayer)
{
  if (PLAYER_STATE_RUN<pPlayer->state)
    PlayerSend(pPlayer,0,PlayerClose);

  if (PLAYER_STATE_NULL<pPlayer->state) {
    ///////////////////////////////////////////////////////////////////////////
    PlayerSend(pPlayer,1,PlayerStop);
    WaitForSingleObject(pPlayer->base.hThread,INFINITE);

    ///////////////////////////////////////////////////////////////////////////
    --pPlayer->state;

    ///////////////////////////////////////////////////////////////////////////
    DPUTS(0,"  destroying thread\n");
    CloseHandle(pPlayer->base.hThread);

#if defined (YASAPI_CHECK_UNDERFLOW) // {
    DPUTS(0,"  destroying timer for underflow\n");
    CloseHandle(pPlayer->base.timer.hUnderflow);
#endif // }

    DPUTS(0,"  destroying timer for read\n");
    CloseHandle(pPlayer->base.timer.hRead);

    DPUTS(0,"  destroying event\n");
    CloseHandle(pPlayer->base.hEvent);

    DPUTS(0,"  destroying store\n");
    StoreDestroy(&pPlayer->base.store);

    DPUTS(0,"  destroying queue\n");
    QueueDestroy(&pPlayer->base.queue);

    DPUTS(0,"  freeing module name\n");
    free(pPlayer->base.pszModuleName);
  }
}

///////////////////////////////////////////////////////////////////////////////
int64_t PlayerRun(Player *pPlayer, Request *pRequest)
{
  IMMDeviceEnumerator *pEnumerator;
  HRESULT hr;

  DPRINTF(0,"  > %s <\n",__func__);

  /////////////////////////////////////////////////////////////////////////////
  if (PLAYER_STATE_BASE!=pPlayer->state) {
    DMESSAGEV("unexpected state: %d",pPlayer->state);
    goto state;
  }

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pPlayer->run,sizeof pPlayer->run);

  // create a device enumerator ///////////////////////////////////////////////
  hr=CoCreateInstance(
    &CLSID_MMDeviceEnumerator,  // _In_   REFCLSID rclsid,
    NULL,                       // _In_   LPUNKNOWN pUnkOuter,
    CLSCTX_ALL,                 // _In_   DWORD dwClsContext,
    &IID_IMMDeviceEnumerator,   // _In_   REFIID riid,
    (void**)&pEnumerator        // _Out_  LPVOID *ppv
  );

  if (FAILED(hr)) {
    DERROR(REGDB_E_CLASSNOTREG,hr);
    DERROR(CLASS_E_NOAGGREGATION,hr);
    DERROR(E_NOINTERFACE,hr);
    DERROR(E_POINTER,hr);
    DMESSAGE("creating device enumerator");
    goto enumerator;
  }

  pPlayer->run.pEnumerator=pEnumerator;
  DPUTS(0,"  device enumerator created\n");

#if defined (YASAPI_NOTIFY) // {
  /////////////////////////////////////////////////////////////////////////////
  if (PlayerAddNotify(pPlayer)<0)
    goto notify;
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  ++pPlayer->state;

  return 0;
//cleanup:
#if defined (YASAPI_NOTIFY) // {
  PlayerRemoveNotify(pPlayer);
notify:
#endif // }
  pEnumerator->lpVtbl->Release(pEnumerator);
enumerator:
state:
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
int64_t PlayerStop(Player *pPlayer, Request *pRequest)
{
  IMMDeviceEnumerator *pEnumerator=pPlayer->run.pEnumerator;

  DPRINTF(0,"  > %s <\n",__func__);

  if (PLAYER_STATE_BASE<pPlayer->state) {
    ///////////////////////////////////////////////////////////////////////////
    --pPlayer->state;

#if defined (YASAPI_NOTIFY) // {
    ///////////////////////////////////////////////////////////////////////////
    DPUTS(0,"  removing notify\n");
    PlayerRemoveNotify(pPlayer);
#endif // }

    DPUTS(0,"  destroying device enumerator\n");
    pEnumerator->lpVtbl->Release(pEnumerator);
  }

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
int PlayerResetTime(Player *pPlayer, int bOffset)
{
  TimeTag eTimeTag=pPlayer->options.device.eTimeTag;
  Time *pCurrent=&pPlayer->current;
  Time *pOffset=&pPlayer->offset;
  Connection *pConnect=&pPlayer->connect;

  if (bOffset) {
    if (TimeReset(pOffset,eTimeTag,pConnect)<0) {
      DMESSAGE("re-setting offset time");
      goto offset;
    }
  }
  else {
    if (TimeGet(pOffset,pConnect)<0) {
      DMESSAGE("getting offset time");
      goto add;
    }
  }

  if (TimeReset(pCurrent,pOffset->eTimeTag,pConnect)<0) {
    DMESSAGE("re-setting current");
    goto current;
  }

  return 0;
offset:
add:
current:
  return -1;
}

int PlayerConnect(Player *pPlayer, int bNegociate, int bReset)
{
  double qShareSize=pPlayer->options.device.qShareSize;
  WAVEFORMATEXTENSIBLE *pwfxx=&pPlayer->open.wfxx;
  WAVEFORMATEX *pwfx=&pwfxx->Format;
  IMMDevice *pDevice=pPlayer->device.pDevice;
  const Strategy *pStrategy;
  AUDCLNT_SHAREMODE eShareMode;
  REFERENCE_TIME hnsDefaultDevicePeriod;
  REFERENCE_TIME hnsMinimumDevicePeriod;
  REFERENCE_TIME hnsDevicePeriod;
  DWORD dwStreamFlags;
  int retry;
  WAVEFORMATEX *pClosestMatch;
  IAudioClient *pClient;
  IAudioRenderClient *pRender;
  IAudioClock *pClock;
  HRESULT hr;

  DPRINTF(0,"  > %s (negociate: %d, reset: %d) <\n",__func__,
      bNegociate,bReset);

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->open.pStrategy=pStrategy
      =pPlayer->options.device.bPull?&gcStrategyPull:&gcStrategyPush;

  if (bNegociate) {
    // setup share mode ///////////////////////////////////////////////////////
    switch (pPlayer->options.device.eShareMode) {
    case YASAPI_SHAREMODE_EXCLUSIVE:
      DPUTS(0,"  YASAPI_SHAREMODE_EXCLUSIVE\n");
      eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
      break;
    case YASAPI_SHAREMODE_AUTOMATIC:
      DPUTS(0,"  YASAPI_SHAREMODE_AUTOMATIC\n");
      eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
      break;
    default:
      DPUTS(0,"  AUDCLNT_SHAREMODE_SHARED\n");
      eShareMode=AUDCLNT_SHAREMODE_SHARED;
      break;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  hnsDevicePeriod=0;
  retry=0;
retry:
  if (bNegociate) {
    // remember the share mode ////////////////////////////////////////////////
    pPlayer->open.eShareMode=eShareMode;
  }
  else
    eShareMode=pPlayer->open.eShareMode;

  // activate the audio client ////////////////////////////////////////////////
  hr=pDevice->lpVtbl->Activate(pDevice,
    &IID_IAudioClient,    // [in]   REFIID iid,
    CLSCTX_ALL,           // [in]   DWORD dwClsCtx,
    NULL,                 // [in]   PROPVARIANT *pActivationParams,
    (void **)&pClient     // [out]  void **ppInterface
  );

  if (FAILED(hr)) {
    DERROR(E_NOINTERFACE,hr);
    DERROR(E_POINTER,hr);
    DERROR(E_INVALIDARG,hr);
    DERROR(E_OUTOFMEMORY,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DUNKNOWN(hr);
    DMESSAGE("activating audio client");
    goto client;
  }

  pPlayer->connect.pClient=pClient;
  DPUTS(0,"  audio client created\n");

  // format supported? ////////////////////////////////////////////////////////
  if (bNegociate&&pPlayer->options.common.bFormatSupported) {
    pClosestMatch=NULL;

    hr=pClient->lpVtbl->IsFormatSupported(pClient,
      eShareMode,         // [in]        AUDCLNT_SHAREMODE ShareMode,
      &pwfxx->Format,     // [in]  const WAVEFORMATEX      *pFormat,
      AUDCLNT_SHAREMODE_EXCLUSIVE==eShareMode?NULL:&pClosestMatch
                          // [out]       WAVEFORMATEX      **ppClosestMatch
    );

    if (pClosestMatch) {
      DPRINTF(0,"  closest match: %d, %d\n",
          pClosestMatch->wBitsPerSample,
          pClosestMatch->nSamplesPerSec);
      CoTaskMemFree(pClosestMatch);
    }

    switch (hr) {
    case S_OK:
      break;
    case S_FALSE:
      DERROR(S_FALSE,hr);
      DMESSAGE("format not supported");
      goto support;
    default:
      if (FAILED(hr)) {
        DERROR(AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
        DERROR(E_POINTER,hr);
        DERROR(E_INVALIDARG,hr);
        DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
        DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
        DUNKNOWN(hr);
        DMESSAGE("format not supported");
        goto support;
      }

      break;
    }

    DPUTS(0,"  format supported\n");
  }

  // setup the stream flags ///////////////////////////////////////////////////
  dwStreamFlags=0;

  if (AUDCLNT_SHAREMODE_SHARED==eShareMode
        &&pPlayer->options.device.bAutoConvertPCM) {
      dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;

     if (pPlayer->options.device.bSRCDefaultQuality)
       dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
  }

  pPlayer->open.dwStreamFlags=dwStreamFlags;

  // if needed, get the minimum device period (low latency) ///////////////////
  // we may have it already because of not aligned buffer error, cf. below.
  if (!bNegociate)
    hnsDevicePeriod=pPlayer->open.hnsDevicePeriod;
  else if (0==hnsDevicePeriod&&pStrategy->NeedDevicePeriod(pPlayer)) {
    hr=pClient->lpVtbl->GetDevicePeriod(pClient,
      &hnsDefaultDevicePeriod,
                          // [out]  REFERENCE_TIME *phnsDefaultDevicePeriod,
      &hnsMinimumDevicePeriod
                          // [out]  REFERENCE_TIME *phnsMinimumDevicePeriod
    );

    if (FAILED(hr)) {
      DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DERROR(E_POINTER,hr);
      DUNKNOWN(hr);
      DMESSAGE("getting device period");
      goto period;
    }

    switch (pPlayer->options.device.eDevicePeriod) {
    case YASAPI_DEVICE_PERIOD_DEFAULT:
      hnsDevicePeriod=hnsDefaultDevicePeriod;
      break;
    case YASAPI_DEVICE_PERIOD_MINIMUM:
      hnsDevicePeriod=hnsMinimumDevicePeriod;
      break;
    default:
      hnsDevicePeriod=hnsMinimumDevicePeriod;
      break;
    }

    DPUTS(0,"  got device period\n");
    DPRINTF(0,"    default: %I64d hns\n",hnsDefaultDevicePeriod);
    DPRINTF(0,"    minimum: %I64d hns\n",hnsMinimumDevicePeriod);
    DPRINTF(0,"    choosen: %I64d hns\n",hnsDevicePeriod);

    if (1.0<qShareSize) {
      hnsDevicePeriod=qShareSize*hnsDevicePeriod+0.5;
      DPRINTF(0,"  device period scaled by %.02f"
          " (device period: %I64d)\n",
          qShareSize,hnsDevicePeriod);
    }

    pPlayer->open.hnsDevicePeriod=hnsDevicePeriod;
  }
  else
    pPlayer->open.hnsDevicePeriod=hnsDevicePeriod;

  // initialize the audio client //////////////////////////////////////////////
  DPUTS(0,"  going to initialize\n");
  DWPRINTF(0,L"    \"%s\"\n",pPlayer->device.szId);
  hr=pPlayer->open.pStrategy->Initialize(pPlayer);

  if (FAILED(hr)) {
    if (bNegociate&&AUDCLNT_SHAREMODE_EXCLUSIVE==eShareMode) {
      switch (hr) {
      case AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED:
        if (0<retry)
          break;

        DPUTS(0,"  failed with AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED\n");
        DPUTS(0,"  trying to get an aligned buffer size\n");
        retry=1;

        // 1. Call IAudioClient::GetBufferSize and receive the
        //    next-highest-aligned buffer size (in frames).
        hr=pClient->lpVtbl->GetBufferSize(pClient,
          &pPlayer->connect.uFrames // [out] UINT32 *pNumBufferFrames
        );

        if (FAILED(hr)) {
          DERROR(AUDCLNT_E_NOT_INITIALIZED,hr);
          DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
          DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
          DERROR(E_POINTER,hr);
          DUNKNOWN(hr);
          DMESSAGE("getting aligned number of frames");
          goto init;
        }

        DPRINTF(0,"  got aligned number of frames: %u\n",
            pPlayer->connect.uFrames);

        // 3. Calculate the aligned buffer size in 100-nansecond units (hns).
        //    The buffer size is (REFERENCE_TIME)((10000.0 * 1000 /
        //    WAVEFORMATEX.nSamplesPerSecond * nFrames) + 0.5). In this
        //    formula, nFrames is the buffer size retrieved by GetBufferSize.
        hnsDevicePeriod=YASAPI_FRAMES_TIMER(pPlayer->connect.uFrames,pwfx)+0.5;

#if 0 // {
        // save hnsDevicePeriod into device options.
        OptionsSetDevicePeriod(pPlayer->options.hnsDevicePeriod);
#endif // }

        // 2. Call IAudioClient::Release to release the audio
        //    client used in the previous call that returned
        //    AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED.
        DPUTS(0,"  destroying audio client\n");
        pClient->lpVtbl->Release(pClient);

        // 4. Call the IMMDevice::Activate method with parameter iid set
        //    to REFIID IID_IAudioClient to create a new audio client.
        // 5. Call Initialize again on the created audio client and
        //    specify the new buffer size and periodicity.
        DPUTS(0,"  retrying with aligned buffer\n");
        goto retry;
      default:
        if (YASAPI_SHAREMODE_AUTOMATIC!=pPlayer->options.device.eShareMode)
          break;

        DPUTS(0,"  initializing audio client failed \n");
        DPUTS(0,"  destroying audio client\n");
        pClient->lpVtbl->Release(pClient);

        DPUTS(0,"  retrying in shared mode\n");
        eShareMode=AUDCLNT_SHAREMODE_SHARED;
        goto retry;
      }
    }

    DERROR(AUDCLNT_E_ALREADY_INITIALIZED,hr);
    DERROR(AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED,hr);
    DERROR(AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(AUDCLNT_E_CPUUSAGE_EXCEEDED,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_DEVICE_IN_USE,hr);
    DERROR(AUDCLNT_E_ENDPOINT_CREATE_FAILED,hr);
    DERROR(AUDCLNT_E_INVALID_DEVICE_PERIOD,hr);
    DERROR(AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
    DERROR(AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED,hr);
    DERROR(AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(E_POINTER,hr);
    DERROR(E_INVALIDARG,hr);
    DERROR(E_OUTOFMEMORY,hr);
    DUNKNOWN(hr);
    DMESSAGE("initializing audio client");
    goto init;
  }

  DPUTS(0,"  audio client initialized\n");

  // set the event handle /////////////////////////////////////////////////////
  hr=pStrategy->SetEvent(pPlayer);

  if (FAILED(hr)) {
    DERROR(E_INVALIDARG,hr);
    DERROR(AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED,hr);
    DERROR(AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(hr);
    DMESSAGE("setting event handle");
    goto event;
  }

  // get the buffer size //////////////////////////////////////////////////////
  //
  // This method retrieves the length of the endpoint buffer shared
  // between the client application and the audio engine. The length
  // is expressed as the number of audio frames the buffer can hold.
  // The size in bytes of an audio frame is calculated as the number
  // of channels in the stream multiplied by the sample size per channel.
  // For example, the frame size is four bytes for a stereo (2-channel)
  // stream with 16-bit samples.
  hr=pClient->lpVtbl->GetBufferSize(pClient,
    &pPlayer->connect.uFrames // [out]  UINT32 *pNumBufferFrames
  );

  if (FAILED(hr)) {
    DERROR(AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(E_POINTER,hr);
    DUNKNOWN(hr);
    DMESSAGE("getting buffer size");
    goto size;
  }

  DPRINTF(0,"  SHARED BUFFER SIZE: %d frames\n",pPlayer->connect.uFrames);
  pPlayer->connect.time=YASAPI_FRAMES_TIMER(pPlayer->connect.uFrames,pwfx);

  // get the render client ////////////////////////////////////////////////////
  hr=pClient->lpVtbl->GetService(pClient,
    &IID_IAudioRenderClient,  // [in]   REFIID riid,
    (void**)&pRender          // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(E_POINTER,hr);
    DERROR(E_NOINTERFACE,hr);
    DERROR(AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(hr);
    DMESSAGE("getting audio render client");
    goto render;
  }

  pPlayer->connect.pRender=pRender;
  DPUTS(0,"  audio render client created\n");

  // get the audio clock //////////////////////////////////////////////////////
  hr=pClient->lpVtbl->GetService(pClient,
    &IID_IAudioClock,         // [in]   REFIID riid,
    (void**)&pClock           // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(E_POINTER,hr);
    DERROR(E_NOINTERFACE,hr);
    DERROR(AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(hr);
    DMESSAGE("getting audio clock");
    goto clock;
  }

  pPlayer->connect.pClock=pClock;
  DPUTS(0,"  audio clock created\n");

  if (bReset) {
    // reset the client ///////////////////////////////////////////////////////
    hr=pClient->lpVtbl->Reset(pClient);

    if (FAILED(hr)) {
      DERROR(AUDCLNT_E_NOT_INITIALIZED,hr);
      DERROR(AUDCLNT_E_NOT_STOPPED,hr);
      DERROR(AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
      DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DUNKNOWN(hr);
      DMESSAGE("re-setting audio client");
      goto reset;
    }

    DPUTS(0,"  audio client reset\n");
  }

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->state=PLAYER_STATE_READY;

  return 0;
reset:
  pClock->lpVtbl->Release(pClock);
  pPlayer->connect.pClock=NULL;
clock:
  pRender->lpVtbl->Release(pRender);
  pPlayer->connect.pRender=NULL;
render:
size:
event:
init:
period:
support:
  pClient->lpVtbl->Release(pClient);
  pPlayer->connect.pClient=NULL;
client:
  return -1;
}

int PlayerCreateWFXX(Player *pPlayer, Request *pRequest)
{
  int samplerate=va_arg(pRequest->ap,int);
  int numchannels=va_arg(pRequest->ap,int);
  int bitspersamp=va_arg(pRequest->ap,int);
  WAVEFORMATEXTENSIBLE *pwfxx=&pPlayer->open.wfxx;
  WAVEFORMATEX *pwfx=&pwfxx->Format;
#if defined (YASAPI_SORROUND) // {
  int bSorround;
#endif // }

  DPRINTF(0,"  > %s (%d, %d, %d) <\n",__func__,
      samplerate,numchannels,bitspersamp);

  /////////////////////////////////////////////////////////////////////////////
  DPRINTF(0,"  samplerate: %d\n",samplerate);

  if (1==numchannels&&pPlayer->options.common.bMono2Stereo) {
    DPUTS(0,"  numchannels: 1 -> 2\n");
    pPlayer->open.nShift=1;
    numchannels=2;
  }
  else {
    DPRINTF(0,"  numchannels: %d\n",numchannels);
    pPlayer->open.nShift=0;
  }

  DPRINTF(0,"  bitspersamp: %d\n",bitspersamp);
#if defined (YASAPI_SORROUND) // {
  bSorround=pPlayer->open.bSorround=pPlayer->options.common.bSorround;
  WFXXSetup(pwfxx,samplerate,numchannels,bitspersamp,bSorround,FALSE);
#else // } {
  WFXXSetup(pwfxx,samplerate,numchannels,bitspersamp,FALSE,FALSE);
#endif // }
  pPlayer->open.wBytesPerSample=pwfx->wBitsPerSample>>3;

  return 0;
state:
  return -1;
}

int PlayerCreateRing(Player *pPlayer)
{
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;
  double qRingFill=pPlayer->options.device.ring.qFill;
  double qRingSize=pPlayer->options.device.ring.qSize;
  UINT32 uBufSize;

  DPRINTF(0,"  > %s <\n",__func__);

  // setup the ring buffer ////////////////////////////////////////////////////
  pPlayer->connect.uFramesMin=qRingFill*pPlayer->connect.uFrames+0.5;
  DPRINTF(0,"  START AUDIO CLIENT: %d frames\n",
      pPlayer->connect.uFramesMin);

  uBufSize=qRingSize*pPlayer->connect.uFrames+0.5;
  // "nBlockAlign" reflects "bMono2Stereo".
  uBufSize*=pwfx->nBlockAlign;

  if (RingCreate(&pPlayer->open.ring,uBufSize)<0) {
    DMESSAGE("creating ring buffer\n");
    goto ring;
  }

  pPlayer->open.ring.Copy=PlayerCopyMemory;
  pPlayer->open.ring.pClient=pPlayer;
#if defined (YASAPI_RING_REALLOC) // {
  pPlayer->open.dwBufSize=uBufSize;
#endif // }
  DPRINTF(0,"  ring buffer created (%d bytes)\n",uBufSize);

  return 0;
//cleanup:
  //RingDestroy(&pPlayer->open.ring);
ring:
  return -1;
}

int PlayerReallocRing(Player *pPlayer)
{
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;
  double qRingFill=pPlayer->options.device.ring.qFill;
  double qRingSize=pPlayer->options.device.ring.qSize;
  UINT32 uBufSize;

  DPRINTF(0,"  > %s <\n",__func__);

  // setup the ring buffer ////////////////////////////////////////////////////
  pPlayer->connect.uFramesMin=qRingFill*pPlayer->connect.uFrames+0.5;
  DPRINTF(0,"  START AUDIO CLIENT: %d frames\n",
      pPlayer->connect.uFramesMin);

  uBufSize=qRingSize*pPlayer->connect.uFrames+0.5;
  // "nBlockAlign" reflects "bMono2Stereo".
  uBufSize*=pwfx->nBlockAlign;

  if (RingRealloc(&pPlayer->open.ring,uBufSize)<0) {
    DMESSAGE("re-allocationg ring buffer\n");
    goto ring;
  }

#if 0 // {
  pPlayer->open.ring.Copy=PlayerCopyMemory;
  pPlayer->open.ring.pClient=pPlayer;
#endif // }
#if defined (YASAPI_RING_REALLOC) // {
  pPlayer->open.dwBufSize=uBufSize;
#endif // }
  DPRINTF(0,"  ring buffer re-allocated (%d bytes)\n",uBufSize);

  return 0;
//cleanup:
  //RingDestroy(&pPlayer->open.ring);
ring:
  return -1;
}

int64_t PlayerGetMaxLatency(Player *pPlayer)
{
  Connection *pConnect=&pPlayer->connect;
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;

  return YASAPI_FRAMES_MS(pConnect->uFramesMin,pwfx);
}

int64_t PlayerOpen(Player *pPlayer, Request *pRequest)
{
  PlayerDevice *pPlayerDevice=va_arg(pRequest->ap,PlayerDevice *);
  const wchar_t *path=pPlayer->options.path;

  DPRINTF(0,"  > %s <\n",__func__);

  /////////////////////////////////////////////////////////////////////////////
  if (pPlayer->state<PLAYER_STATE_RUN) {
    DMESSAGEV("unexpected state: %d (%s)",pPlayer->state,__func__);
    goto state;
  }

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pPlayer->device,sizeof pPlayer->device);
  ZeroMemory(&pPlayer->open,sizeof pPlayer->open);
  ZeroMemory(&pPlayer->connect,sizeof pPlayer->connect);
#if defined (YASAPI_GAPLESS) // {
  ZeroMemory(&pPlayer->offset,sizeof pPlayer->offset);
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->device=*pPlayerDevice;
  ZeroMemory(pPlayerDevice,sizeof *pPlayerDevice);
  pPlayerDevice=&pPlayer->device;

  if (!pPlayerDevice->pDevice
      &&PlayerDeviceGet(pPlayerDevice,pPlayer->run.pEnumerator)<0) {
    DMESSAGE("getting the device");
    goto device;
  }

  DPUTS(0,"  got the device\n");

  /////////////////////////////////////////////////////////////////////////////
  OptionsDeviceLoad(&pPlayer->options.device,L"    ",
      pPlayerDevice->szId,path);
  DPUTS(0,"  device options loaded\n");

  // Create WAVEFORMATEXTENSIBLE //////////////////////////////////////////////
  if (PlayerCreateWFXX(pPlayer,pRequest)<0) {
    DMESSAGE("creating WAVEFORMATEXTENSIBLE");
    goto wfxx;
  }

  // Connect //////////////////////////////////////////////////////////////////
  if (PlayerConnect(pPlayer,1,1)<0) {
    DMESSAGE("connecting ");
    goto connect;
  }

  // Create ring //////////////////////////////////////////////////////////////
  if (PlayerCreateRing(pPlayer)<0) {
    DMESSAGE("creating ring");
    goto ring;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (PlayerResetTime(pPlayer,1)<0) {
    DMESSAGE("re-setting gapless offset");
    goto time;
  }

  return PlayerGetMaxLatency(pPlayer);
time:
  RingDestroy(&pPlayer->open.ring);
ring:
connect:
wfxx:
  PlayerDeviceDestroy(pPlayerDevice);
device:
state:
  pPlayer->state=PLAYER_STATE_RUN;
  return -1;
}

#if defined (YASAPI_NOTIFY) // {
int64_t PlayerMigrate(Player *pPlayer, Request *pRequest)
{
  const wchar_t *path=pPlayer->options.path;
  PlayerDevice device;

  DPRINTF(0,"  > %s <\n",__func__);

  ////////////////////////////////////////////////////////////////////////////
  if (pPlayer->state<PLAYER_STATE_PAUSE) {
    DPRINTF(0,"  Warning: illegal state: %d.\n",pPlayer->state);
    goto state;
  }

  ////////////////////////////////////////////////////////////////////////////
  if (PlayerDeviceCreate(&device,L"",pPlayer->run.pEnumerator)<0) {
    DPUTS(0,"  Warning: getting the default device");
    goto device1;
  }

  if (0==wcscmp(pPlayer->device.szId,device.szId)) {
    DPUTS(0,"  Warning: device not changed");
    goto device2;
  }

  if (PlayerDeviceGet(&device,pPlayer->run.pEnumerator)<0) {
    DMESSAGE("  Warning: getting the device");
    goto device3;
  }

  DPUTS(0,"  got the device\n");

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->connect.bInvalid=1;
  PlayerKill(pPlayer,1,PLAYER_STATE_RUN);
  
  pPlayer->device=device;
  ZeroMemory(&device,sizeof device);

  /////////////////////////////////////////////////////////////////////////////
  OptionsDeviceLoad(&pPlayer->options.device,
      L"    ",
      pPlayer->device.szId,
      path);
  DPUTS(0,"  device options loaded\n");

  // Connect //////////////////////////////////////////////////////////////////
  if (PlayerConnect(pPlayer,1,1)<0) {
    DMESSAGE("connecting ");
    goto connect;
  }

  // Realloc ring /////////////////////////////////////////////////////////////
  if (PlayerReallocRing(pPlayer)<0) {
    DMESSAGE("re-allocating ring");
    goto ring;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (PlayerResetTime(pPlayer,0)<0) {
    DMESSAGE("re-setting gapless offset");
    goto time;
  }

  pPlayer->state=PLAYER_STATE_READY;
  PlayerTryStart(pPlayer);

  return PlayerGetMaxLatency(pPlayer);
time:
  RingDestroy(&pPlayer->open.ring);
ring:
connect:
  PlayerDeviceDestroy(&pPlayer->device);
device3:
device2:
device1:
  pPlayer->state=PLAYER_STATE_RUN;
state:
  return -1;
}
#endif // }

void PlayerShutdown(Player *pPlayer)
{
  IAudioClient *pClient=pPlayer->connect.pClient;

  DPRINTF(0,"  %s: cancelling timer\n",__func__);
  CancelWaitableTimer(pPlayer->base.timer.hRead);

  if (!pPlayer->connect.bInvalid) {
    DPRINTF(0,"  %s: stopping audio client\n",__func__);
    pClient->lpVtbl->Stop(pClient);
  }

  PlayerPostUpdate(pPlayer,0,0);
}

void PlayerDisconnect(Player *pPlayer)
{
  IAudioClient *pClient=pPlayer->connect.pClient;
  IAudioRenderClient *pRender=pPlayer->connect.pRender;
  IAudioClock *pClock=pPlayer->connect.pClock;

  if (pPlayer->connect.bInvalid||pPlayer->options.common.bDisconnect) {
    if (NULL!=pClock) {
      DPUTS(0,"  destroying audio clock\n");
      pClock->lpVtbl->Release(pClock);
    }

    if (NULL!=pRender) {
      DPUTS(0,"  destroying render client\n");
      pRender->lpVtbl->Release(pRender);
    }

    if (NULL!=pClient&&!pPlayer->connect.bInvalid) {
      DPUTS(0,"  destroying audio client\n");
      pClient->lpVtbl->Release(pClient);
    }

    ZeroMemory(&pPlayer->connect,sizeof pPlayer->connect);
  }
}

void PlayerKill(Player *pPlayer, int bReset, const PlayerState state)
{
  IAudioClient *pClient=pPlayer->connect.pClient;
  HRESULT hr;

  DPRINTF(0,"  > %s (%d) <\n",__func__,state);

  switch (pPlayer->state) {
  case PLAYER_STATE_EOT:
    if (state==pPlayer->state)
      break;

    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_DRAIN_3:
    if (state==pPlayer->state)
      break;

    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_DRAIN_2:
    if (state==pPlayer->state)
      break;

    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_DRAIN_1:
    if (state==pPlayer->state)
      break;

    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_PLAY:
    if (state==pPlayer->state)
      break;

    PlayerShutdown(pPlayer);
    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_READY:
    if (state==pPlayer->state)
      break;

    if (pClient&&bReset&&!pPlayer->connect.bInvalid) {
      if (PlayerResetTime(pPlayer,0)<0)
        DPRINTF(0,"  Warning: re-setting time failed.\n");

      hr=pClient->lpVtbl->Reset(pClient);

      if (FAILED(hr)) {
        DERROR(AUDCLNT_E_NOT_INITIALIZED,hr);
        DERROR(AUDCLNT_E_NOT_STOPPED,hr);
        DERROR(AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
        DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
        DUNKNOWN(hr);
        DPRINTF(0,"  Warning: re-setting audio client failed.\n");
      }
      else
        DPRINTF(0,"  audio client reset.\n");
    }

    PlayerDisconnect(pPlayer);
    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_UNDERFLOW:
    if (state==pPlayer->state)
      break;

    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_PAUSE:
    if (state==pPlayer->state)
      break;

    PlayerDeviceDestroy(&pPlayer->device);
    --pPlayer->state;
    // intentional fall throuh, no break!
  default:
    break;
  }
}

#if defined (YASAPI_GAPLESS) // {
int64_t PlayerReset(Player *pPlayer, Request *pRequest)
{
  Connection *pConnect=&pPlayer->connect;
  IAudioClient *pClient=pConnect->pClient;
  WAVEFORMATEXTENSIBLE *pwfxx=&pPlayer->open.wfxx;
  WAVEFORMATEX *pFormat=&pwfxx->Format;
  Time *pOffset=&pPlayer->offset;
  UINT32 u32FramesPadding;
  UINT32 u32FramesWritten;
  UINT32 u32FramesWaiting;

  DPRINTF(0,"  > %s <\n",__func__);
  DPUTS(0,"  re-using previous device\n");

  if (pPlayer->state<PLAYER_STATE_PAUSE) {
    DPRINTF(0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }

  if (pConnect->bInvalid)
    goto invalid;

  if (pPlayer->state<PLAYER_STATE_READY&&PlayerConnect(pPlayer,0,0)<0) {
    DMESSAGE("connecting");
    goto connect;
  }

  if (PlayerResetTime(pPlayer,0)<0) {
    DMESSAGE("re-setting gapless offset\n");
    goto reset;
  }

  if ((u32FramesPadding=PlayerGetFramesPadding(pPlayer))<0) {
    DMESSAGE("getting frames padding");
    goto padding;
  }

  // get the number of frames written to the ring buffer
  u32FramesWritten=pPlayer->open.ring.dwWritten/pFormat->nBlockAlign;
  // get the number of frames padding and written to the ring buffer
  u32FramesWaiting=u32FramesPadding+u32FramesWritten;

  if (TimeAddFrames(pOffset,u32FramesWaiting,pFormat,pConnect)<0) {
    DMESSAGE("adding frames");
    goto add;
  }

  return PlayerGetMaxLatency(pPlayer);
add:
padding:
reset:
connect:
invalid:
state:
  return -1;
}
#endif // }

int64_t PlayerClose(Player *pPlayer, Request *pRequest)
{
  PlayerDevice *pPlayerDevice=&pPlayer->device;

  DPRINTF(0,"  > %s <\n",__func__);

  if (PLAYER_STATE_RUN<pPlayer->state) {
    PlayerKill(pPlayer,1,PLAYER_STATE_RUN);
    DPUTS(0,"  destroying ring buffer\n");
    RingDestroy(&pPlayer->open.ring);
  }

  return 0;
}

int64_t PlayerCanWrite(Player *pPlayer, Request *pRequest)
{
  enum { DEBUG=3 };
#if defined (YASAPI_DEBUG) // {
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;
#endif // }
  DWORD dwAvailable;

  DPRINTF(DEBUG,"  > %s (%d) <\n",__func__,pPlayer->state);

  if (pPlayer->state<PLAYER_STATE_PAUSE) {
    DPRINTF(0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    dwAvailable=0;
  }
#if defined (YASAPI_NOTIFY) // {
  else if (pPlayer->connect.bInvalid) {
    DPUTS(DEBUG,"  invalid connection\n");
    dwAvailable=0;
  }
#endif // }
  else if (pPlayer->state<PLAYER_STATE_UNDERFLOW) {
    DPUTS(DEBUG,"  pause\n");
    dwAvailable=0;
  }
  else {
#if defined (YASAPI_RING_REALLOC) // {
    DPRINTF(DEBUG,"  buf size: %d, written: %d\n",
        pPlayer->open.dwBufSize,pPlayer->open.ring.dwWritten);

    if (pPlayer->open.dwBufSize<pPlayer->open.ring.dwWritten)
      dwAvailable=0;
    else
      dwAvailable=pPlayer->open.dwBufSize-pPlayer->open.ring.dwWritten;
#else // } {
    dwAvailable=pPlayer->open.ring.dwAvailable;
#endif // }
    dwAvailable>>=pPlayer->open.nShift;
  }

  return dwAvailable;
}

//#define YASAPI_AUDCLNT_E_BUFFER_SIZE_ERROR
int64_t PlayerPlay(Player *pPlayer, UINT32 uFramesPadding)
{
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;
  int bNeedPadding=pPlayer->open.pStrategy->NeedPadding(pPlayer);
  IAudioRenderClient *pRender=pPlayer->connect.pRender;
  Ring *pRing=&pPlayer->open.ring;
#if defined (YASAPI_AUDCLNT_E_BUFFER_SIZE_ERROR) // {
  UINT32 uFramesWrite=pPlayer->connect.uFrames-uFramesPadding;
  DWORD dwSize=uFramesWrite*pwfx->nBlockAlign;
  DWORD dwWrite=pPlayer->open.ring.dwWritten;
#else // } {
  UINT32 uFramesWrite;
  DWORD dwSize;
  DWORD dwWrite;
#endif // }
  DWORD dwWriteSize;
  DWORD dwFlags=0;
#if defined (YASAPI_RING_DWORD) // {
  DWORD dwFramesRead;
#else // } {
  UINT32 uFramesRead;
#endif // }
  BYTE *pData;
  WORD wRing,wShared;
  HRESULT hr;

  DPRINTF(3,"  > %s (%d) <\n",__func__,pPlayer->state);

  /////////////////////////////////////////////////////////////////////////////
  if (pPlayer->state<PLAYER_STATE_READY) {
    DPRINTF(0,"Warning: unexpected state: %d: \"%s\" (%d).\n",
        &pPlayer->state,basename(__FILE__),__LINE__);
    PlayerKill(pPlayer,1,PLAYER_STATE_RUN);
    goto state;
  }

#if ! defined (YASAPI_AUDCLNT_E_BUFFER_SIZE_ERROR) // {
  uFramesWrite=pPlayer->connect.uFrames;

  if (!bNeedPadding)
  	uFramesWrite-=uFramesPadding;

  dwSize=uFramesWrite*pwfx->nBlockAlign;
  dwWrite=pPlayer->open.ring.dwWritten;
#endif // }

  if (dwWrite<dwSize) {
    if (pPlayer->state<PLAYER_STATE_DRAIN_1) {
      if (bNeedPadding||0==(dwWriteSize=dwWrite)) {
        DPRINTF(0,"  %s: underflow\n",__func__);
        PlayerKill(pPlayer,0,PLAYER_STATE_UNDERFLOW);
        goto underflow;
      }
    }

	dwWriteSize=dwWrite;

    if (!bNeedPadding)
      uFramesWrite=dwWriteSize/pwfx->nBlockAlign;
  }
  else
    dwWriteSize=dwSize;

  if (PlayerGetDlgConfig(pPlayer)) {
    wRing=MulDiv(USHRT_MAX,pRing->dwWritten,pRing->pMax-pRing->pRep);
    wShared=MulDiv(USHRT_MAX,uFramesPadding,pPlayer->connect.uFrames);
    PlayerPostUpdate(pPlayer,wRing,wShared);
  }

  if (0==uFramesWrite)
    goto nop;

  if (uFramesWrite*pwfx->nBlockAlign!=dwWriteSize) {
    DMESSAGE("uFrameWrite does not match dwWriteSize");
    goto match;
  }

  // get the render buffer ////////////////////////////////////////////////////
  hr=pRender->lpVtbl->GetBuffer(pRender,
    uFramesWrite,     // [in]   UINT32 NumFramesRequested,
    &pData            // [out]  BYTE **ppData
  );

  if (FAILED(hr)) {
#if defined (YASAPI_NOTIFY) // {
    if (AUDCLNT_E_DEVICE_INVALIDATED==hr) {
      pPlayer->connect.bInvalid=1;
      goto invalid;
    }
#endif // }

    DERROR(AUDCLNT_E_BUFFER_ERROR,hr);
    DERROR(AUDCLNT_E_BUFFER_TOO_LARGE,hr);
    DERROR(AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(E_POINTER,hr);
    DUNKNOWN(hr);
    DMESSAGE("getting render buffer");
    goto get;
  }

  // copy the data ////////////////////////////////////////////////////////////
#if defined (YASAPI_RING_DWORD) // {
  dwFramesRead=RingReadEx(pRing,pData,dwWriteSize,RING_COPY);

  if (0<dwFramesRead)
    dwFramesRead/=pwfx->nBlockAlign;
#else // } {
  uFramesRead=RingReadEx(pRing,pData,dwWriteSize,RING_COPY);
  uFramesRead/=pwfx->nBlockAlign;
#endif // }

  if (dwWriteSize<dwSize&&bNeedPadding)
    ZeroMemory(pData+dwWriteSize,dwSize-dwWriteSize);

  // release the render buffer ////////////////////////////////////////////////
#if defined (YASAPI_RING_DWORD) // {
  hr=pRender->lpVtbl->ReleaseBuffer(pRender,
    0<dwFramesRead?dwFramesRead:0,
                      // [in]  UINT32 NumFramesWritten,
    dwFlags           // [in]  DWORD dwFlags
  );
#else // } {
  hr=pRender->lpVtbl->ReleaseBuffer(pRender,
    uFramesRead,      // [in]  UINT32 NumFramesWritten,
    dwFlags           // [in]  DWORD dwFlags
  );
#endif // }

  if (FAILED(hr)) {
    DERROR(AUDCLNT_E_INVALID_SIZE,hr);
    DERROR(AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(E_INVALIDARG,hr);
    DUNKNOWN(hr);
    DMESSAGE("releasing render buffer");
    goto release;
  }

#if defined (YASAPI_RING_DWORD) // {
  if (dwFramesRead<0) {
    DMESSAGE("reading ring buffer");
    goto read;
  }
#endif // }

  // no critical section needed in order to block unexpected write
  // requests to the ring buffer because all requests are serialized.
  RingReadEx(pRing,pData,dwWriteSize,RING_COMMIT);
nop:
  if (PLAYER_STATE_DRAIN_1==pPlayer->state&&0==pPlayer->open.ring.dwWritten)
    ++pPlayer->state;

  pPlayer->open.pStrategy->SetTimer(pPlayer,pPlayer->connect.time,FALSE);
#if defined (YASAPI_NOTIFY) // {
invalid:
#endif // }
underflow:
state:
  return 0;
#if defined (YASAPI_RING_DWORD) // {
read:
#endif // }
release:
get:
match:
  return -1;
}

int PlayerTryStart(Player *pPlayer)
{
  const Strategy *pStrategy=pPlayer->open.pStrategy;
  const PlayerState state=pPlayer->state;
  IAudioClient *pClient=pPlayer->connect.pClient;
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;
  Ring *pRing=&pPlayer->open.ring;
  UINT32 uFramesRingWritten=pRing->dwWritten/pwfx->nBlockAlign;
  UINT32 uFramesPadding;
  HRESULT hr;

  DPRINTF(0,"  > %s (%d) <\n",__func__,pPlayer->state);

  if (pPlayer->state<PLAYER_STATE_UNDERFLOW) {
    DPRINTF(0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }

  if (pPlayer->state<PLAYER_STATE_PLAY) {
    if ((uFramesPadding=pStrategy->GetFramesPadding(pPlayer))<0) {
      DMESSAGE("getting frames padding");
      goto padding;
    }

    if (uFramesRingWritten+uFramesPadding<pPlayer->connect.uFramesMin) {
      DPRINTF(0,"  %s: underflow\n",__func__);
      goto underflow;
    }

    // start the audio client /////////////////////////////////////////////////
    hr=pClient->lpVtbl->Start(pClient);
  
    if (FAILED(hr)) {
      DERROR(AUDCLNT_E_NOT_INITIALIZED,hr);
      DERROR(AUDCLNT_E_NOT_STOPPED,hr);
      DERROR(AUDCLNT_E_EVENTHANDLE_NOT_SET,hr);
      DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DUNKNOWN(hr);
      DMESSAGE("starting audio client");
      goto start;
    }

    DPRINTF(0,"  %s: audio client started\n",__func__);

    ///////////////////////////////////////////////////////////////////////////
    pPlayer->state=PLAYER_STATE_PLAY;

    ///////////////////////////////////////////////////////////////////////////
    if (PlayerPlay(pPlayer,uFramesPadding)<0) {
      DMESSAGEV("playing");
      goto play;
    }

    DPRINTF(0,"  %s: start playing\n",__func__);
  }
underflow:
  return 0;
play:
start:
padding:
  pPlayer->state=state;
state:
  return -1;
}

int64_t PlayerWrite(Player *pPlayer, Request *pRequest)
{
  const char *buf=va_arg(pRequest->ap,const char *);
  int len=va_arg(pRequest->ap,int);
  Ring *pRing=&pPlayer->open.ring;

  DPRINTF(3,"  > %s (%d) <\n",__func__,pPlayer->state);

  if (pPlayer->state<PLAYER_STATE_PAUSE) {
    DPRINTF(0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }

  if (PLAYER_STATE_PAUSE==pPlayer->state)
    goto pause;

#if defined (YASAPI_RING_REALLOC) // {
  if (RingReallocAvailable(pRing,len,pPlayer->open.nShift)<0) {
    DMESSAGE("reallocating");
    goto realloc;
  }
#endif // }

#if defined (YASAPI_RING_DWORD) // {
  if (RingWrite(pRing,buf,len,pPlayer->open.nShift)<0) {
    DMESSAGE("writing");
    goto write;
  }
#else // } {
  RingWrite(pRing,buf,len,pPlayer->open.nShift);
#endif // }

  if (pPlayer->state<PLAYER_STATE_PLAY)
    PlayerTryStart(pPlayer);
pause:
  return 0;
#if defined (YASAPI_RING_DWORD) // {
write:
#endif // }
#if defined (YASAPI_RING_REALLOC) // {
realloc:
#endif // }
//ring:
state:
  return -1;
}

int64_t PlayerFlush(Player *pPlayer, Request *pRequest)
{
  const PlayerState state=pPlayer->state;
  IAudioClient *pClient=pPlayer->connect.pClient;
  HRESULT hr;
  int64_t time;

  DPRINTF(0,"  > %s (%d) <\n",__func__,pPlayer->state);

  if (pPlayer->state<PLAYER_STATE_PAUSE) {
    DPRINTF(0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }

  if (pPlayer->connect.bInvalid)
    goto invalid;

  RingReset(&pPlayer->open.ring);
  
  if (PLAYER_STATE_UNDERFLOW<pPlayer->state) {
    PlayerKill(pPlayer,1,PLAYER_STATE_READY);

    if (PlayerResetTime(pPlayer,0)<0) {
      DMESSAGE("re-setting gapless offset\n");
      goto time;
    }

    if ((time=PlayerGetTime(pPlayer,NULL))<0) {
      DMESSAGE("getting time\n");
      goto get;
    }

    return time;
  }
  else
    return 0;
get:
time:
invalid:
state:
  return 0;
}

int64_t PlayerIsPlaying(Player *pPlayer, Request *pRequest)
{
  DPRINTF(3,"  > %s (%d) <\n",__func__,pPlayer->state);

  if (pPlayer->state<PLAYER_STATE_PAUSE) {
    DPRINTF(0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    return 0;
  }
  else if (pPlayer->open.ring.dwWritten) {
    pPlayer->state=PLAYER_STATE_DRAIN_1;
    DPUTS(0,"  isplaying\n");
    return 1;
  }
  else {
    switch (pPlayer->state) {
    case PLAYER_STATE_PAUSE:
    case PLAYER_STATE_UNDERFLOW:
    case PLAYER_STATE_READY:
      pPlayer->state=PLAYER_STATE_EOT;
      DPUTS(0,"  isplaying (eot)\n");
      return 0;
    case PLAYER_STATE_PLAY:
      pPlayer->state=PLAYER_STATE_DRAIN_2;
      DPUTS(0,"  isplaying\n");
      return 1;
    case PLAYER_STATE_DRAIN_1:
    case PLAYER_STATE_DRAIN_2:
    case PLAYER_STATE_DRAIN_3:
      DPRINTF(0,"  isplaying (drain %d)\n",
          1+pPlayer->state-PLAYER_STATE_DRAIN_1);
      return 1;
    case PLAYER_STATE_EOT:
    default:
      DPUTS(0,"  isplaying (eot)\n");
      return 0;
    }
  }
}


int64_t PlayerPause(Player *pPlayer, Request *pRequest)
{
  int bPause=PLAYER_STATE_PAUSE==pPlayer->state;
	int64_t time=0;

  DPRINTF(0,"  > %s (%d) <\n",__func__,pPlayer->state);

  if (pPlayer->state<PLAYER_STATE_PAUSE) {
    DPRINTF(0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }

  if (bPause) {
#if defined (YASAPI_NOTIFY) // {
    if (pPlayer->connect.bInvalid)
      time=0;
    else {
#endif // }
      if (PlayerConnect(pPlayer,0,0)<0) {
        DMESSAGE("connecting");
        goto connect;
      }

      if ((time=PlayerGetTime(pPlayer,NULL))<0) {
        DMESSAGE("getting time");
        goto get1;
      }

      PlayerTryStart(pPlayer);
#if defined (YASAPI_NOTIFY) // {
    }
#endif // }
  }
  else {
#if defined (YASAPI_NOTIFY) // {
    if (pPlayer->connect.bInvalid)
      time=0;
    else {
#endif // }
      if ((time=PlayerGetTime(pPlayer,NULL))<0) {
        DMESSAGE("getting time");
        goto get2;
      }

      if (PlayerResetTime(pPlayer,0)<0) {
        DMESSAGE("re-setting gapless offset");
        goto time;
      }

      PlayerKill(pPlayer,0,PLAYER_STATE_PAUSE);
#if defined (YASAPI_NOTIFY) // {
    }
#endif // }
  }

  return time;
time:
get2:
get1:
connect:
state:
  return -1;
}

int64_t PlayerGetTime(Player *pPlayer, Request *pRequest)
{
  Time *pOffset=&pPlayer->offset;
  Time *pCurrent=&pPlayer->current;
  Connection *pConnect=&pPlayer->connect;
  double qTime;

  DPRINTF(3,"  > %s (%d) <\n",__func__,pPlayer->state);

  if (pPlayer->state<PLAYER_STATE_READY) {
    DPRINTF(3,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }

  if (pPlayer->connect.bInvalid)
    goto invalid;

  if (TimeInSeconds(pOffset,pCurrent,pConnect,&qTime)<0) {
    DMESSAGE("getting time");
    goto time;
  }

  return 1000.0*qTime+0.5;
time:
invalid:
state:
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_RING_REALLOC) // {
void PlayerWriteArgs(Player *pPlayer, Result *pResult, ...)
{
  Request request;
  va_list ap;
  int state;

  ZeroMemory(&request, sizeof request);
  va_start(ap,pResult);
  request.pResult=pResult;
  request.ap=ap;
  pResult->state=PlayerWrite(pPlayer,&request);
  va_end(ap);
}
#endif // }

int64_t PlayerRead(Player *pPlayer, Request *pRequest)
{
  UINT32 uFramesPadding;

  DPRINTF(3,"  > %s (%d) <\n",__func__,pPlayer->state);

  if (pPlayer->state<PLAYER_STATE_READY) {
    DPRINTF(0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }
#if defined (YASAPI_NOTIFY) // {
  else if (pPlayer->connect.bInvalid)
    goto invalid;
#endif // }
  else if (PLAYER_STATE_DRAIN_2==pPlayer->state) {
    pPlayer->state=PLAYER_STATE_DRAIN_3;
    pPlayer->open.pStrategy->SetTimer(pPlayer,pPlayer->connect.time,TRUE);
  }
  else if (PLAYER_STATE_DRAIN_3==pPlayer->state) {
    pPlayer->state=PLAYER_STATE_EOT;
    goto eof;
  }
  else if (PLAYER_STATE_PAUSE==pPlayer->state)
    goto pause;
  else {
    if ((uFramesPadding=PlayerGetFramesPadding(pPlayer))<0) {
      DMESSAGE("getting frames padding");
      goto padding;
    }

    if (PlayerPlay(pPlayer,uFramesPadding)<0) {
      DMESSAGE("playing");
      goto play;
    }
  }
pause:
eof:
  return 0;
play:
padding:
#if defined (YASAPI_NOTIFY) // {
invalid:
#endif // }
state:
  return -1;
}

#if defined (YASAPI_CHECK_UNDERFLOW) // {
int PlayerShutdownUnderflow(Player *pPlayer, Request *pRequest)
{
  if (PLAYER_STATE_UNDERFLOW==pPlayer->state)
    PlayerClose(pPlayer,pRequest);

  return 0;
}
#endif // }

///////////////////////////////////////////////////////////////////////////////
UINT32 PlayerGetFramesPadding(Player *pPlayer)
{
  IAudioClient *pClient=pPlayer->connect.pClient;
  UINT32 uFramesPadding;
  HRESULT hr;

#if defined (YASAPI_NOTIFY) // {
  if (pPlayer->connect.bInvalid)
    goto invalid;
#endif // }

  hr=pClient->lpVtbl->GetCurrentPadding(pClient,
    &uFramesPadding       // [out]  UINT32 *pNumPaddingFrames
  );

  if (FAILED(hr)) {
#if defined (YASAPI_NOTIFY) // {
    if (AUDCLNT_E_DEVICE_INVALIDATED==hr) {
      pPlayer->connect.bInvalid=1;
      goto invalid;
    }
#endif // }

    DERROR(AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(E_POINTER,hr);
    DUNKNOWN(hr);
    DMESSAGE("getting padding");
    goto padding;
  }

  return uFramesPadding;
padding:
#if defined (YASAPI_NOTIFY) // {
invalid:
#endif // }
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
HWND PlayerGetDlgConfig(Player *pPlayer)
{
  return pPlayer->options.common.bVisualization?pPlayer->hDlgConfig:NULL;
}

void PlayerPostUpdate(Player *pPlayer, WORD wRing, WORD wShared)
{
  HWND hDlgConfig=PlayerGetDlgConfig(pPlayer);

  if (hDlgConfig)
    PostMessage(hDlgConfig,WM_CONFIG_UPDATE,0,MAKELONG(wRing,wShared));
}

///////////////////////////////////////////////////////////////////////////////
int64_t PlayerSend(Player *pPlayer, int exit, PlayerProc *pPlayerProc, ...)
{
  Store *pStore=&pPlayer->base.store;
  Queue *pQueue=&pPlayer->base.queue;
  Result *pResult;
  Request *pRequest;
  int state;
  va_list ap;

  va_start(ap,pPlayerProc);
  pResult=StoreGet(pStore);
  pRequest=QueueLockWrite(pQueue,pResult,exit,pPlayerProc,pPlayer,ap);
  QueueUnlockWrite(pQueue,pResult);
    state=pResult->state;
  StorePut(pStore,pResult);
  va_end(ap);

  return state;
}

void PlayerPost(Player *pPlayer, int exit, PlayerProc *pPlayerProc)
{
  Queue *pQueue=&pPlayer->base.queue;

  QueueLockWrite(pQueue,NULL,exit,pPlayerProc,pPlayer,NULL);
  QueueUnlockWrite(pQueue,NULL);
}

///////////////////////////////////////////////////////////////////////////////
void PlayerPostRead(void *data)
{
  PlayerPost(data,0,PlayerRead);
}

VOID CALLBACK PlayerFireRead(LPVOID lpArg, DWORD dwLow, DWORD dwHigh)
{
  PlayerPostRead(lpArg);
}

void PlayerSetTimerRead(Player *pPlayer, LONGLONG time)
{
  HANDLE hTimer=pPlayer->base.timer.hRead;
  LARGE_INTEGER ui;

  // http://msdn.microsoft.com/en-us/library/windows/desktop/ms687012%28v=vs.85%29.aspx
  // Positive values indicate absolute time. 
  // Negative values indicate relative time.
  // in 100 nanosecond intervals
  ui.QuadPart=-time;

  SetWaitableTimer(
    hTimer,           // _In_      HANDLE hTimer,
    &ui,              // _In_      const LARGE_INTEGER *pDueTime,
    0,                // _In_      LONG lPeriod,
    time?PlayerFireRead:NULL,
                      // _In_opt_  PTIMERAPCROUTINE pfnCompletionRoutine,
    time?pPlayer:NULL,
                      // _In_opt_  LPVOID lpArgToCompletionRoutine,
    FALSE             // _In_      BOOL fResume
  );
}

#if defined (YASAPI_CHECK_UNDERFLOW) // {
///////////////////////////////////////////////////////////////////////////////
VOID CALLBACK PlayerFireShutdownUnderflow(LPVOID lpArg, DWORD dwLow,
    DWORD dwHigh)
{
  PlayerPost(lpArg,0,PlayerShutdownUnderflow);
}

void PlayerSetTimerCheckUnderflow(Player *pPlayer)
{
  int ms=pPlayer->options.common.nCheckUnderflow;
  HANDLE hTimer=pPlayer->base.timer.hUnderflow;
  LARGE_INTEGER ui;

  // http://msdn.microsoft.com/en-us/library/windows/desktop/ms687012%28v=vs.85%29.aspx
  // Positive values indicate absolute time. 
  // Negative values indicate relative time.
  // in 100 nanosecond intervals
  ui.QuadPart=-1000*1000*ms/100;

  SetWaitableTimer(
    hTimer,           // _In_      HANDLE hTimer,
    &ui,              // _In_      const LARGE_INTEGER *pDueTime,
    0,                // _In_      LONG lPeriod,
    ms?PlayerFireShutdownUnderflow:NULL,
                      // _In_opt_  PTIMERAPCROUTINE pfnCompletionRoutine,
    ms?pPlayer:NULL,
                      // _In_opt_  LPVOID lpArgToCompletionRoutine,
    FALSE             // _In_      BOOL fResume
  );
}
#endif // }

///////////////////////////////////////////////////////////////////////////////
HANDLE PlayerGetEvent(Player *pPlayer)
{
  return pPlayer->state<PLAYER_STATE_PLAY
      ?NULL:pPlayer->open.pStrategy->GetEvent(pPlayer);
}

DWORD WINAPI PlayerThread(LPVOID lpParameter)
{
  Player *pPlayer=lpParameter;
  Queue *pQueue=&pPlayer->base.queue;
  int exit=0;
  DWORD dwTaskIndex=0;
  HANDLE hTask=AvSetMmThreadCharacteristicsW(L"Pro Audio",&dwTaskIndex);
  HANDLE hEvent;
  Request *pRequest;
  int state;
#if defined (YASAPI_THREAD_CO_INITIALIZE) // {
  HRESULT hr;

  hr=CoInitializeEx(
    NULL,                     // _In_opt_ LPVOID pvReserved
	  COINIT_MULTITHREADED		  // _In_     DWORD  dwCoInit
  );

  if (FAILED(hr)) {
    DERROR(S_FALSE,hr);
    DERROR(RPC_E_CHANGED_MODE,hr);
    DMESSAGE("initializing com failed");
    goto coinit;
  }
#endif // }

  while (!exit) {
    hEvent=PlayerGetEvent(pPlayer);
    pRequest=QueueLockRead(pQueue,hEvent,PlayerPostRead,pPlayer);
    state=pRequest->pPlayerProc(pPlayer,pRequest);
    exit=pRequest->exit;

    if (NULL!=pRequest->pResult)
      pRequest->pResult->state=state;

    QueueUnlockRead(&pPlayer->base.queue,pRequest->pResult);
  }

#if defined (YASAPI_THREAD_CO_INITIALIZE) // {
  CoUninitialize();
coinit:
#endif // }

  if (hTask)
    AvRevertMmThreadCharacteristics(hTask);

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
void PlayerCopyMemory(PVOID p, PVOID Destination, const VOID *Source,
    SIZE_T Length)
{
  Player *pPlayer=p;
  Options *pOptions=&pPlayer->options;
  double qVolume=pPlayer->base.qVolume;
  char *wp,*mp;
  const char *rp;
  int m,n;
  double q;
  int32_t i32;

  wp=Destination;
  mp=wp+Length;
  rp=Source;
  // bits per sample.
  n=pPlayer->open.wfxx.Format.wBitsPerSample;
  // bytes per sample.
  m=n>>3;
  // attenuation factor.
  q=qVolume*YASAPI_INT_MAX(n)/YASAPI_INT_MAX(32);
  // bits to shift the 32-bit result.
  n=32-n;

  if (!pPlayer->open.nShift) {
    if (!pOptions->common.bVolume||1.0==qVolume) {
//fprintf(stderr,"  %s 1\n",__func__);
      CopyMemory(
        Destination,            // _In_  PVOID Destination,
        Source,                 // _In_  const VOID *Source,
        Length                  // _In_  SIZE_T Length
      );
//fprintf(stderr,"  %s 2\n",__func__);
    }
    else {
//fprintf(stderr,"  %s 3\n",__func__);
      while (wp<mp) {
        i32=0;

        CopyMemory(
          &i32,                 // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        i32<<=n;
        i32=q*i32+0.5;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          &i32,                 // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;
        rp+=m;
      }
//fprintf(stderr,"  %s 4\n",__func__);
    }
  }
  else {
    if (!pOptions->common.bVolume||1.0==qVolume) {
//fprintf(stderr,"  %s 5\n",__func__);
      while (wp<mp) {
        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;
        rp+=m;
      }
//fprintf(stderr,"  %s 6\n",__func__);
    }
    else {
//fprintf(stderr,"  %s 7\n",__func__);
      while (wp<mp) {
        i32=0;

        CopyMemory(
          &i32,                 // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        i32<<=n;
        i32=q*i32+0.5;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          &i32,                 // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          &i32,                 // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;
        rp+=m;
      }
//fprintf(stderr,"  %s 8\n",__func__);
    }
  }
}
