/*
 * yasapi_queue.c
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

// win32 thread-safe queue implementation using native windows API
// http://stackoverflow.com/questions/11706985/win32-thread-safe-queue-implementation-using-native-windows-api

///////////////////////////////////////////////////////////////////////////////
int QueueCreate(Queue *pQueue)
{
  /////////////////////////////////////////////////////////////////////////////
  pQueue->wp=pQueue->rp=pQueue->aRequest;
  pQueue->mp=pQueue->aRequest+QUEUE_SIZE;

  /////////////////////////////////////////////////////////////////////////////
  pQueue->hAvailable=CreateSemaphore(
    0,            // _In_opt_ LPSECURITY_ATTRIBUTES lpSemaphoreAttributes,
    QUEUE_SIZE,   // _In_     LONG                  lInitialCount,
    QUEUE_SIZE,   // _In_     LONG                  lMaximumCount,
    NULL          // _In_opt_ LPCTSTR               lpName
  );

  if (NULL==pQueue->hAvailable) {
    DMESSAGE("creating available semaphore");
    goto available;
  }
  
  DPUTS(0,"  queue available semaphore created\n");

  /////////////////////////////////////////////////////////////////////////////
  pQueue->hWritten=CreateSemaphore(
    0,            // _In_opt_ LPSECURITY_ATTRIBUTES lpSemaphoreAttributes,
    0,            // _In_     LONG                  lInitialCount,
    QUEUE_SIZE,   // _In_     LONG                  lMaximumCount,
    NULL          // _In_opt_ LPCTSTR               lpName
  );

  if (NULL==pQueue->hWritten) {
    DMESSAGE("creating written semaphore");
    goto written;
  }
  
  DPUTS(0,"  queue written semaphore created\n");

  /////////////////////////////////////////////////////////////////////////////
  pQueue->hMutex=CreateMutex(
    NULL,         // _In_opt_ LPSECURITY_ATTRIBUTES lpMutexAttributes,
    FALSE,        // _In_     BOOL                  bInitialOwner,
    NULL          // _In_opt_ LPCTSTR               lpName
  );

  if (NULL==pQueue->hMutex) {
    DMESSAGE("creating mutex");
    goto mutex;
  }
  
  DPUTS(0,"  queue mutex created\n");

#if defined (YASAPI_DEBUG) // {
  /////////////////////////////////////////////////////////////////////////////
  pQueue->nMinAvailable=pQueue->nAvailable=QUEUE_SIZE;
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
// cleanup:
  //CloseHandle(pQueue->hMutex);
mutex:
  CloseHandle(pQueue->hWritten);
written:
  CloseHandle(pQueue->hAvailable);
available:
  return -1;
}

void QueueDestroy(Queue *pQueue)
{
  DPUTS(0,"  destroying queue mutex\n");
  CloseHandle(pQueue->hMutex);
  DPUTS(0,"  destroying queue written semaphore\n");
  CloseHandle(pQueue->hWritten);
  DPUTS(0,"  destroying queue available semaphore\n");
  CloseHandle(pQueue->hAvailable);
}

///////////////////////////////////////////////////////////////////////////////
void QueueLockMutex(Queue *pQueue)
{
  WaitForSingleObjectEx(
    pQueue->hMutex,   // _In_ HANDLE hHandle,
    INFINITE,         // _In_ DWORD  dwMilliseconds,
    FALSE             // _In_ BOOL   bAlertable
  );
}

void QueueUnlockMutex(Queue *pQueue)
{
  ReleaseMutex(pQueue->hMutex);
}

///////////////////////////////////////////////////////////////////////////////
Request *QueueLock(Queue *pQueue, const QueueStrategy *pStrategy,
    HANDLE hEvent, void (*OnEvent)(void *data), void *data)
{
  BOOL bAlertable=pStrategy->IsAlertable(pQueue);
  DWORD nCount=0;
  HANDLE ahHandles[2];
  DWORD dwCode;
  Request *pRequest;

  ahHandles[nCount++]=pStrategy->GetSemaphoreDown(pQueue);

  if (hEvent)
    ahHandles[nCount++]=hEvent;
retry:
  dwCode=WaitForMultipleObjectsEx(
    nCount,           // _In_       DWORD  nCount,
    ahHandles,        // _In_ const HANDLE *lpHandles,
    FALSE,            // _In_       BOOL   bWaitAll,
    INFINITE,         // _In_       DWORD  dwMilliseconds,
    bAlertable        // _In_       BOOL   bAlertable
  );

  switch (dwCode) {
  case WAIT_IO_COMPLETION:
    // alertable event (timer): we need to loop.
    goto retry;
  case WAIT_OBJECT_0+0:
    break;
  case WAIT_OBJECT_0+1:
    if (OnEvent)
      OnEvent(data);

    // event: we need to loop.
    goto retry;
  default:
    break;
  }

  QueueLockMutex(pQueue);
#if defined (YASAPI_DEBUG) // {
  pStrategy->Dec(pQueue);
#endif // }

  pRequest=*pStrategy->GetRequest(pQueue);
  return pRequest;
}

void QueueUnlock(Queue *pQueue, const QueueStrategy *pStrategy,
    Result *pResult)
{
  Request **ppRequest;

  ppRequest=pStrategy->GetRequest(pQueue);

  if (++*ppRequest==pQueue->mp)
    *ppRequest=pQueue->aRequest;

  if (pResult)
    SetEvent(pResult->hEvent);

#if defined (YASAPI_DEBUG) // {
  pStrategy->Inc(pQueue);
#endif // }
  QueueUnlockMutex(pQueue);
  ReleaseSemaphore(pStrategy->GetSemaphoreUp(pQueue),1,NULL);
}

Request *QueueLockWrite(Queue *pQueue, Result *pResult, int exit,
    PlayerProc *pPlayerProc, Player *pPlayer, va_list ap)
{
  Request *pRequest;

  pRequest=QueueLock(pQueue,&cqsWrite,NULL,NULL,NULL);
  pRequest->pResult=pResult;
  pRequest->exit=exit;
  pRequest->pPlayerProc=pPlayerProc;
  pRequest->ap=ap;

  return pRequest;
}

void QueueUnlockWrite(Queue *pQueue, Result *pResult)
{
  QueueUnlock(pQueue,&cqsWrite,NULL);

  if (pResult) {
    WaitForSingleObjectEx(
      pResult->hEvent,      // _In_ HANDLE hHandle,
      INFINITE,             // _In_ DWORD  dwMilliseconds,
      FALSE                 // _In_ BOOL   bAlertable
    );
  }
}

Request *QueueLockRead(Queue *pQueue, HANDLE hEvent,
    void (*OnEvent)(void *data), void *data)
{
  Request *pRequest;

  pRequest=QueueLock(pQueue,&cqsRead,hEvent,OnEvent,data);

  return pRequest;
}

void QueueUnlockRead(Queue *pQueue, Result *pResult)
{
  QueueUnlock(pQueue,&cqsRead,pResult);
}

void QueueWrite(Queue *pQueue, Store *pStore, int exit,
    PlayerProc *pPlayerProc, Player *pPlayer, va_list ap)
{
  Result *pResult;
  Request *pRequest;

  pResult=pStore?StoreGet(pStore):NULL;
  pRequest=QueueLockWrite(pQueue,pResult,exit,pPlayerProc,pPlayer,ap);
  QueueUnlockWrite(pQueue,pResult);
  StorePut(pStore,pResult);
}

#if defined (YASAPI_DEBUG) // {
///////////////////////////////////////////////////////////////////////////////
void QueueIncAvailable(Queue *pQueue)
{
  ++pQueue->nAvailable;
}

void QueueDecAvailable(Queue *pQueue)
{
  if (--pQueue->nAvailable<pQueue->nMinAvailable)
    pQueue->nMinAvailable=pQueue->nAvailable;
}

void QueueNopAvailable(Queue *pQueue)
{
}
#endif // }
