/*
 * yasapi_ring.c
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#include <stdint.h>

#if defined (YASAPI_DEBUG) // {
#define YASAPI_RING_DEBUG
#endif // }

#if defined (YASAPI_RING_DEBUG) // {
#define RING_CONSISTENCY(pRing,dwSize,func,line,label) \
  if (RingConsistency(pRing,dwSize,func,line)<0) { \
    DMESSAGEV("%s: consistency",__func__); \
    goto label; \
  }
#else // } {
#define RING_CONSISTENCY(pRing,dwSize,func,line,label)
#endif // }

DWORD RingCreate(Ring *pRing, SIZE_T dwSize)
{
  pRing->Copy=RingCopyMemory;
  pRing->pClient=pRing;

  // create ring buffer.
  if (NULL==(pRing->pRep=YASAPI_MALLOC(dwSize))) {
    DMESSAGE("allocating ring buffer");
    goto rep;
  }
  
  pRing->pMax=pRing->pRep+dwSize;
  RingReset(pRing);

  return dwSize;
// cleanup:
  //YASAPI_FREE(pRing->pRep);
rep:
  return -1;
}

void RingDestroy(Ring *pRing)
{
  if (pRing->pRep)
    YASAPI_FREE(pRing->pRep);
}

void RingReset(Ring *pRing)
{
  pRing->pHead=pRing->pRep;
  pRing->pTail=pRing->pRep;
  pRing->dwWritten=0;
  pRing->dwAvailable=pRing->pMax-pRing->pRep;
}

static int RingConsistency(Ring *pRing, DWORD dwSize,
    const char *func, int line)
{
  DWORD dwWritten=pRing->dwWritten+dwSize;
  DWORD dwAvailable=pRing->dwAvailable-dwSize;
  DWORD dwBufSize=pRing->pMax-pRing->pRep;
  DWORD dwOffs1;
  DWORD dwOffs2;
  DWORD dwDiff;

  if (pRing->pTail<=pRing->pHead) {
    dwOffs1=pRing->pTail-pRing->pRep;
    dwOffs2=pRing->pHead-pRing->pRep;
    dwDiff=dwOffs2-dwOffs1;

    if (0==dwDiff&&(dwBufSize==dwWritten||0==dwAvailable))
      ;
    else if (dwWritten==dwDiff)
      ;
#if 0 // {
    else if (0==(dwOffs2-dwOffs1)&&0==dwAvailable)
      ;
#endif // }
    else {
      DMESSAGEV("%s: un-wrapped (%d)",func,line);
      goto unwrapped;
    }
#if 0 // {
    if (dwWritten!=dwOffs2-dwOffs1) {
      DMESSAGEV("%s: %d %d un-wrapped (%d)\n",func,
          dwWritten,dwOffs2-dwOffs1,line);
      goto unwrapped;
    }
#endif // }
  }
  else {
    dwOffs1=pRing->pHead-pRing->pRep;
    dwOffs2=pRing->pMax-pRing->pTail;

    if (dwWritten!=dwOffs1+dwOffs2) {
      DMESSAGEV("%s: wrapped (%d)",func,line);
      goto wrapped;
    }
  }

  return 0;
size:
unwrapped:
wrapped:
  return -1;
}

#if defined (YASAPI_RING_REALLOC) // {
int RingRealloc(Ring *pRing, SIZE_T uSize)
{
  UINT32 uBufSize=pRing->pMax-pRing->pRep;
#if defined (YASAPI_DEBUG) // {
  int debug;
#endif // }
  DWORD dwAvailable;
  DWORD dwOffs1;
  DWORD dwOffs2;
  LPSTR pRep;
  LPSTR pMax;
  LPSTR pHead;
  LPSTR pTail;

  if (uBufSize<uSize) {
    RING_CONSISTENCY(pRing,0,__func__,__LINE__,consistency1);

    do {
      if (0==(uBufSize<<=1)) {
        DMESSAGEV("%s: overflow",__func__);
          goto overflow;
      }
    } while (uBufSize<uSize);

    uSize=uBufSize;
    dwAvailable=uSize-pRing->dwWritten;

    DPRINTF(0,"  > %s (%d bytes) <\n",__func__,uSize);

    if (NULL==(pRep=(LPSTR)YASAPI_MALLOC(uSize))) {
      DMESSAGEV("%s: malloc",__func__);
      goto malloc;
    }

    pMax=pRep+uSize;

    // are we un-wrapped?
    if (pRing->pTail<pRing->pHead) {
      // yes, we are.
      dwOffs1=pRing->pTail-pRing->pRep;
      dwOffs2=pRing->pHead-pRing->pRep;

      pTail=pRep+dwOffs1;
      pHead=pRep+dwOffs2;

      if (0<pRing->dwWritten) {
        CopyMemory(
          pTail,              // _In_       PVOID  Destination,
          pRing->pTail,       // _In_ const VOID   *Source,
          pRing->dwWritten    // _In_       SIZE_T Length
        );
      }
    }
    else if (pRing->pHead<pRing->pTail||0<pRing->dwWritten) {
      // no, we are wrapped.
      dwOffs1=pRing->pHead-pRing->pRep;
      dwOffs2=pRing->pMax-pRing->pTail;

      pHead=pRep+dwOffs1;

      if (0<dwOffs1) {
        CopyMemory(
          pRep,               // _In_       PVOID  Destination,
          pRing->pRep,        // _In_ const VOID   *Source,
          dwOffs1             // _In_       SIZE_T Length
        );
      }

      pTail=pMax-dwOffs2;

      if (0<dwOffs2) {
        CopyMemory(
          pTail,              // _In_       PVOID  Destination,
          pRing->pTail,       // _In_ const VOID   *Source,
          dwOffs2             // _In_       SIZE_T Length
        );
      }
    }
    else {
      pTail=pRep;
      pHead=pRep;
    }

    YASAPI_FREE(pRing->pRep);
    pRing->pRep=pRep;
    pRing->pMax=pMax;
    pRing->pTail=pTail;
    pRing->pHead=pHead;
    pRing->dwAvailable=dwAvailable;

    RING_CONSISTENCY(pRing,0,__func__, __LINE__,consistency2);

    return 0;
  }
  else
    return 0;
consistency2:
  YASAPI_FREE(pRep);
malloc:
overflow:
consistency1:
  return -1;
}

int RingReallocAvailable(Ring *pRing, SIZE_T dwAvailable, SIZE_T n)
{
  return RingRealloc(pRing,pRing->dwWritten+dwAvailable);
}
#endif // }

#if defined (YASAPI_RING_DWORD) // {
static DWORD RingWriteUnwrapped(Ring *pRing, LPCSTR pData, SIZE_T dwSize,
    SIZE_T nShift)
{
  if (0<dwSize) {
    pRing->Copy(
      pRing->pClient,
      pRing->pHead,     // _In_  PVOID Destination,
      pData,            // _In_  const VOID *Source,
      dwSize            // _In_  SIZE_T Length
    );

    pData+=dwSize>>nShift;
    pRing->pHead+=dwSize;
  }

  RING_CONSISTENCY(pRing,dwSize,__func__, __LINE__,consistency);

  return dwSize;
consistency:
  return -1;
}

static DWORD RingWriteWrapped(Ring *pRing, LPCSTR pData, SIZE_T dwSize,
    SIZE_T nShift)
{
  DWORD dwOffs1=pRing->pMax-pRing->pHead;
  DWORD dwOffs2=dwSize-dwOffs1;

  if (0<dwOffs1) {
    pRing->Copy(
      pRing->pClient,
      pRing->pHead,     // _In_  PVOID Destination,
      pData,            // _In_  const VOID *Source,
      dwOffs1           // _In_  SIZE_T Length
    );

    pData+=dwOffs1>>nShift;
    pRing->pHead+=dwOffs2;
  }

  if (0<dwOffs2) {
    pRing->Copy(
      pRing->pClient,
      pRing->pRep,      // _In_  PVOID Destination,
      pData,            // _In_  const VOID *Source,
      dwOffs2           // _In_  SIZE_T Length
    );

    pData+=dwOffs2>>nShift;
    pRing->pHead=pRing->pRep+dwOffs2;
  }

  RING_CONSISTENCY(pRing,dwSize,__func__, __LINE__,consistency);

  return dwSize;
consistency:
  return -1;
}

DWORD RingWrite(Ring *pRing, LPCSTR pData, SIZE_T dwSize, SIZE_T nShift)
{
  RING_CONSISTENCY(pRing,0,__func__, __LINE__,consistency1);

  dwSize<<=nShift;

  if (pRing->dwAvailable<dwSize)
    dwSize=pRing->dwAvailable;

  if (0<dwSize) {
    if (pRing->pTail==pRing->pHead) {
      if (0==pRing->dwWritten) {
        if (pRing->pMax<pRing->pHead+dwSize) {
          if (RingWriteWrapped(pRing,pData,dwSize,nShift)<0)
            goto wrapped;
        }
        else {
          if (RingWriteUnwrapped(pRing,pData,dwSize,nShift)<0)
            goto unwrapped;
        }
      }
      else {
        DMESSAGEV("%s: overflow",__func__);
        goto overflow;
      }
    }
    else if (pRing->pTail<pRing->pHead) {
      if (pRing->pMax<pRing->pHead+dwSize) {
        if (RingWriteWrapped(pRing,pData,dwSize,nShift)<0)
          goto wrapped;
      }
      else {
        if (RingWriteUnwrapped(pRing,pData,dwSize,nShift)<0)
          goto unwrapped;
      }
    }
    else {
      if (RingWriteUnwrapped(pRing,pData,dwSize,nShift)<0)
        goto wrapped;
    }

    pRing->dwWritten+=dwSize;
    pRing->dwAvailable-=dwSize;
  }

  RING_CONSISTENCY(pRing,0,__func__, __LINE__,consistency2);

  return dwSize;
consistency2:
overflow:
unwrapped:
wrapped:
consistency1:
  return -1;
}
#else // } {
SIZE_T RingWrite(Ring *pRing, LPCSTR pData, SIZE_T dwSize, SIZE_T nShift)
{
  SIZE_T dwCopySize;

  dwSize<<=nShift;

  if (dwSize>pRing->dwAvailable)
    dwSize=pRing->dwAvailable;

  if (0<dwSize) {
    if (0<(pRing->pHead-pRing->pTail)||dwSize<=(pRing->pMax-pRing->pTail)) {
      if (NULL!=pData) {
        pRing->Copy(
          pRing->pClient,
          pRing->pTail,       // _In_  PVOID Destination,
          pData,              // _In_  const VOID *Source,
          dwSize              // _In_  SIZE_T Length
        );
      }

      pRing->pTail+=dwSize;
    }
    else {
      dwCopySize=pRing->pMax-pRing->pTail;

      if (0<dwCopySize&&NULL!=pData) {
        pRing->Copy(
          pRing->pClient,
          pRing->pTail,       // _In_  PVOID Destination,
          pData,              // _In_  const VOID *Source,
          dwCopySize          // _In_  SIZE_T Length
        );

        pData+=dwCopySize>>nShift;
      }

      dwCopySize=dwSize-dwCopySize;

      if (0<dwCopySize&&NULL!=pData) {
        pRing->Copy(
          pRing->pClient,
          pRing->pRep,        // _In_  PVOID Destination,
          pData,              // _In_  const VOID *Source,
          dwCopySize          // _In_  SIZE_T Length
        );
      }

      pRing->pTail=pRing->pRep+dwCopySize;
    }

    pRing->dwWritten+=dwSize;
    pRing->dwAvailable-=dwSize;
  }

  return dwSize;
}
#endif // }

#if defined (YASAPI_RING_DWORD) // {
static DWORD RingReadUnwrappedEx(Ring *pRing, LPSTR pData, SIZE_T dwSize,
    UINT uFlags)
{
  if (RING_COPY&uFlags&&0<dwSize) {
    CopyMemory(
      pData,          // _In_  PVOID Destination,
      pRing->pTail,     // _In_  const VOID *Source,
      dwSize            // _In_  SIZE_T Length
    );

    pData+=dwSize;
  }

  if (RING_COMMIT&uFlags) {
    pRing->pTail+=dwSize;
    RING_CONSISTENCY(pRing,-dwSize,__func__, __LINE__,consistency);
  }

  return dwSize;
consistency:
  return -1;
}

static DWORD RingReadWrappedEx(Ring *pRing, LPSTR pData, SIZE_T dwSize,
    UINT uFlags)
{
  DWORD dwOffs1=pRing->pMax-pRing->pTail;
  DWORD dwOffs2=pRing->pHead-pRing->pRep;
  DWORD dwCopySize;

  dwCopySize=dwSize<dwOffs1?dwSize:dwOffs1;

  if (RING_COPY&uFlags) {
    CopyMemory(
      pData,              // _In_  PVOID Destination,
      pRing->pTail,       // _In_  const VOID *Source,
      dwCopySize          // _In_  SIZE_T Length
    );

    pData+=dwCopySize;
  }

  if (RING_COMMIT&uFlags) {
    pRing->pTail+=dwCopySize;
    RING_CONSISTENCY(pRing,-dwCopySize,__func__, __LINE__,consistency1);
  }

  // is there somthing left to read?
  if (dwCopySize<dwSize) {
    if (dwOffs2<(dwCopySize=dwSize-dwCopySize)) {
      DMESSAGEV("%s: overread",__func__);
      goto overread;
    }

    if (RING_COPY&uFlags) {
      CopyMemory(
        pData,            // _In_  PVOID Destination,
        pRing->pRep,      // _In_  const VOID *Source,
        dwCopySize        // _In_  SIZE_T Length
      );

      pData+=dwCopySize;
    }
  
    // we are not wrapped any longer.
    if (RING_COMMIT&uFlags) {
      pRing->pTail=pRing->pRep+dwCopySize;
      RING_CONSISTENCY(pRing,-dwSize,__func__, __LINE__,consistency2);
    }
  }

  return dwSize;
consistency2:
overread:
consistency1:
  return -1;
}

DWORD RingReadEx(Ring *pRing, LPSTR pData, SIZE_T dwSize, UINT uFlags)
{
  RING_CONSISTENCY(pRing,0,__func__, __LINE__,consistency1);

  if (pRing->dwWritten<dwSize)
    dwSize=pRing->dwWritten;

  if (0<dwSize) {
    if (pRing->pTail==pRing->pHead) {
      if (0==pRing->dwAvailable) {
        if (pRing->pMax<pRing->pTail+dwSize) {
          if (RingReadWrappedEx(pRing,pData,dwSize,uFlags)<0)
            goto wrapped;
        }
        else {
          if (RingReadUnwrappedEx(pRing,pData,dwSize,uFlags)<0)
            goto unwrapped;
        }
      }
      else {
        DMESSAGEV("%s: overread",__func__);
        goto overread;
      }
    }
    else if (pRing->pTail<pRing->pHead) {
      if (pRing->pMax<pRing->pTail+dwSize) {
        if (RingReadWrappedEx(pRing,pData,dwSize,uFlags)<0)
          goto wrapped;
      }
      else {
        if (RingReadUnwrappedEx(pRing,pData,dwSize,uFlags)<0)
          goto unwrapped;
      }
    }
    else {
      if (RingReadWrappedEx(pRing,pData,dwSize,uFlags)<0)
        goto wrapped;
    }

    if (RING_COMMIT&uFlags) {
      pRing->dwWritten-=dwSize;
      pRing->dwAvailable+=dwSize;
    }
  }

  RING_CONSISTENCY(pRing,0,__func__, __LINE__,consistency2);

  return dwSize;
consistency2:
overread:
unwrapped:
wrapped:
consistency1:
  return -1;
}

DWORD RingRead(Ring *pRing, LPSTR pData, SIZE_T dwSize)
{
  return RingReadEx(pRing,pData,dwSize,RING_COPY|RING_COMMIT);
}
#else // } {
SIZE_T RingReadEx(Ring *pRing, LPSTR pData, SIZE_T dwSize, UINT uFlags)
{
  SIZE_T dwCopySize;

  if (dwSize>pRing->dwWritten)
    dwSize=pRing->dwWritten;

  if (0<dwSize) {
    if (0<(pRing->pTail-pRing->pHead)||dwSize<=(pRing->pMax-pRing->pHead)) {
      if (NULL!=pData&&(RING_COPY&uFlags)) {
        CopyMemory(
          pData,              // _In_  PVOID Destination,
          pRing->pHead,       // _In_  const VOID *Source,
          dwSize              // _In_  SIZE_T Length
        );
      }

      if (RING_COMMIT&uFlags)
        pRing->pHead+=dwSize;
    }
    else {
      dwCopySize=pRing->pMax-pRing->pHead;

      if (0<dwCopySize&&NULL!=pData) {
        if (RING_COPY&uFlags) {
          CopyMemory(
            pData,            // _In_  PVOID Destination,
            pRing->pHead,     // _In_  const VOID *Source,
            dwCopySize        // _In_  SIZE_T Length
          );
        }

        pData+=dwCopySize;
      }

      dwCopySize=dwSize-dwCopySize;

      if (0<dwCopySize&&NULL!=pData&&(RING_COPY&uFlags)) {
        CopyMemory(
          pData,              // _In_  PVOID Destination,
          pRing->pRep,        // _In_  const VOID *Source,
          dwCopySize          // _In_  SIZE_T Length
        );
      }

      if (RING_COMMIT&uFlags)
        pRing->pHead=pRing->pRep+dwCopySize;
    }

    if (RING_COMMIT&uFlags) {
      pRing->dwAvailable+=dwSize;
      pRing->dwWritten-=dwSize;
    }
  }

  return dwSize;
}

SIZE_T RingRead(Ring *pRing, LPSTR pData, SIZE_T dwSize)
{
  return RingReadEx(pRing,pData,dwSize,RING_COPY|RING_COMMIT);
}
#endif // }

void RingCopyMemory(PVOID *Client, PVOID Destination, const VOID *Source,
    SIZE_T Length)
{
  CopyMemory(
    Destination,              // _In_  PVOID Destination,
    Source,                   // _In_  const VOID *Source,
    Length                    // _In_  SIZE_T Length
  );
}
