/*
 * yasapi_time.c
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

INT64 ConnectionGetPosition(Connection *pConnect)
{
  IAudioClock *pClock=pConnect->pClock;
  UINT64 u64Position;
  UINT64 u64QPCPosition;
  HRESULT hr;

  if (pConnect->bInvalid)
    goto invalid;

  // get position /////////////////////////////////////////////////////////////
  hr=pClock->lpVtbl->GetPosition(pClock,
    &u64Position,         // [out] UINT64 *pu64Position,
    &u64QPCPosition       // [out] UINT64 *pu64QPCPosition
  );

  if (FAILED(hr)) {
    if (hr==AUDCLNT_E_DEVICE_INVALIDATED) {
      pConnect->bInvalid=1;
      goto invalid;
    }

    DERROR(E_POINTER,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(hr);
    DMESSAGE("getting position");
    goto position;
  }

  return u64Position;
position:
invalid:
  return -1;
}

INT64 ConnectionGetFrequency(Connection *pConnect)
{
  IAudioClock *pClock=pConnect->pClock;
  UINT64 u64Frequency;
  HRESULT hr;

  if (pConnect->bInvalid)
    goto invalid;

  // get frequency ////////////////////////////////////////////////////////////
  hr=pClock->lpVtbl->GetFrequency(pClock,
    &u64Frequency         // [out] UINT64 *pu64Frequency
  );

  if (FAILED(hr)) {
    if (hr==AUDCLNT_E_DEVICE_INVALIDATED) {
      pConnect->bInvalid=1;
      goto invalid;
    }

    DERROR(E_POINTER,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(hr);
    DMESSAGE("getting frequency");
    goto frequency;
  }

  return u64Frequency;
frequency:
invalid:
  return -1;
}

int TimeReset(Time *pTime, TimeTag eTimeTag, Connection *pConnect)
{
  enum { DEBUG=3 };
  INT64 i64Frequency;

  DPRINTF(DEBUG,"  > %s <\n",__func__);

  if (pConnect->bInvalid)
    goto invalid;

  pTime->eTimeTag=eTimeTag;

  switch (pTime->eTimeTag) {
  case TIME_POSITION:
    if ((i64Frequency=ConnectionGetFrequency(pConnect))<=0) {
      DMESSAGE("getting frequency");
      goto frequency;
    }

    pTime->u64Position=0ull;
    pTime->u64Frequency=i64Frequency;
    DPRINTF(DEBUG,"  time reset to position %I64u\n",pTime->u64Position);

    break;
  case TIME_TIME:
    pTime->qTime=0.0;

    DPRINTF(DEBUG,"  time reset to time\n");

    break;
  default:
    DMESSAGEV("tag mismatch: %d\n",pTime->eTimeTag);
    goto tag;
  }

  return 0;
tag:
frequency:
invalid:
  return -1;
}

int TimeGet(Time *pOffset, Connection *pConnect)
{
  enum { DEBUG=3 };
  INT64 i64Position;
  INT64 i64Frequency;

  DPRINTF(DEBUG,"  > %s <\n",__func__);

  if (pConnect->bInvalid)
    goto invalid;

  if ((i64Position=ConnectionGetPosition(pConnect))<0) {
    DMESSAGE("getting position");
    goto position;
  }

  switch (pOffset->eTimeTag) {
  case TIME_POSITION:
    pOffset->u64Position=i64Position;
    DPRINTF(DEBUG,"  time set to position %I64u\n",pOffset->u64Position);

    break;
  case TIME_TIME:
    if (i64Frequency=ConnectionGetFrequency(pConnect)<=0) {
      DMESSAGE("getting frequency");
      goto frequency;
    }

    pOffset->qTime=(double)i64Position/i64Frequency;
    DPRINTF(DEBUG,"  time set to time\n");

    break;
  default:
    DMESSAGEV("tag mismatch: %d\n",pOffset->eTimeTag);
    goto tag;
  }

  return 0;
tag:
frequency:
position:
invalid:
  return -1;
}

int TimeAddFrames(Time *pTime, int nFrames, WAVEFORMATEX *pFormat,
    Connection *pConnect)
{
  enum { DEBUG=3 };
  INT64 i64Frequency;
  INT64 i64PositionFrames;

  DPRINTF(DEBUG,"  > %s <\n",__func__);

  if (pConnect->bInvalid)
    goto invalid;

  // get the frequency
  switch (pTime->eTimeTag) {
  case TIME_POSITION:
    i64Frequency=pTime->u64Frequency;
    break;
  case TIME_TIME:
    if ((i64Frequency=ConnectionGetFrequency(pConnect))<=0) {
      DMESSAGE("getting frequency");
      goto frequency;
    }

    break;
  default:
    DMESSAGEV("tag mismatch: %d\n",pTime->eTimeTag);
    goto tag;
  }

  // convert frames to position
  i64PositionFrames=MulDiv(nFrames*pFormat->nChannels,
      i64Frequency,pFormat->nAvgBytesPerSec);

  switch (pTime->eTimeTag) {
  case TIME_POSITION:
    pTime->u64Position+=i64PositionFrames;
    DPRINTF(DEBUG,"  position offset: %I64d\n",pTime->u64Position);
    break;
  case TIME_TIME:
    pTime->qTime+=(double)i64PositionFrames/i64Frequency;
    DPRINTF(DEBUG,"  time offset: %f\n",pTime->qTime);
    break;
  default:
    DMESSAGEV("tag mismatch: %d\n",pTime->eTimeTag);
    goto tag;
  }

  return 0;
tag:
frequency:
position:
invalid:
  return -1;
}

int TimeInSeconds(const Time *pOffset, Time *pCurrent, Connection *pConnect,
    double *pTime)
{
  enum { DEBUG=3 };
  INT64 i64Position;
  INT64 i64Frequency;

  DPRINTF(DEBUG,"  > %s <\n",__func__);

  if (pConnect->bInvalid)
    goto invalid;

  if (pOffset->eTimeTag!=pCurrent->eTimeTag) {
    DMESSAGE("different time tags in get");
    goto tag;
  }

  if ((i64Position=ConnectionGetPosition(pConnect))<0) {
    DMESSAGE("getting position");
    goto position;
  }

  switch (pOffset->eTimeTag) {
  case TIME_POSITION:
    pCurrent->u64Position=i64Position;
    *pTime=pCurrent->u64Position-pOffset->u64Position;
    *pTime/=pOffset->u64Frequency;
    break;
  case TIME_TIME:
    if ((i64Frequency=ConnectionGetFrequency(pConnect))<=0) {
      DMESSAGE("getting frequency");
      goto frequency;
    }

    pCurrent->qTime=(double)i64Position/i64Frequency;
    *pTime=pCurrent->qTime-pOffset->qTime;
    break;
  default:
    DMESSAGEV("tag mismatch: %d\n",pOffset->eTimeTag);
    goto tag;
  }

  return 0;
frequency:
position:
tag:
invalid:
  return -1;
}
